package logic;
import java.util.*;

/** 
 *  This class is responsible for deciding which algorithm to 
 *  use and whether consistency checking should be done.
 *  It uses the if-then-else framework to perform the logic,
 *  using the new algorithm directly. The Rules class implements
 *  Updateable so that the two values that are determined -- the
 *  algorithm class and the consistency check flag -- can be
 *  directly set into Rules, which in turn reads these and 
 *  executes the appropriate loadRules function.
 */
public class AlgorithmAndConsisCheckSelector {
    
    /** performance testing shows that, on a machine with specs:
     *  Win98, 400 mhz processor, 384 mb ram, the
     *  framework performance degrades when consistency checking is
     *  on and the number of rules reaches around 1000
     */
    final int NUM_RULES_THRESHOLD = 1000;
    public static final String ALGORITHM="algorithm";
    public static final String CONSIS_CHECK = "consis_check";
    public static final String EXCEPTION = "exception";
    
    /** 
     *  These are the rules specified by the calling application.
     *  In this class, the internal rules that are used for deciding
     *  the proper algorithm are names "rules".
     */
    Rules outerRules;
    
    AlgorithmAndConsisCheckSelector(Rules rules) {
        this.outerRules = rules;
    }
    
    /**
     *  This static inner class encapsulates the table data that
     *  resulted from performance testing, which indicates which
     *  algorithm performs faster when consistency checking is on.
     *  It is never to be consulted if consistency checking is off.
     */
    public static class _Table implements Updateable, Cloneable {
        
        final int [] POWERS_OF_2 = {512,1024,2048,4096,8192,16384}; 
        final int MAX_RULES = POWERS_OF_2[(POWERS_OF_2.length-1)];
        /** last index where any average number of wildcards is acceptable, and many-wildcard algortithm
         *  could be used
         */
        final int LAST_INDEX = 2;
        final int MAX_AVG_NUM_WILDCARDS = 4;
        Boolean F = Boolean.FALSE;
        Boolean T = Boolean.TRUE;
        
                                           //  avg num wildcards        
                                           // 0   1   2     3     4
        Boolean [][] useManyWildcardAlg = {
         /*index determined by  */ 
         /*the number of rules  */  /*0:*/  { T,  T,  T,    T,    T },
                                    /*1:*/  { F,  T,  T,    T,    T },
                                    /*2:*/  { F,  F,  F,    T,    T },
                                    /*3:*/  { F,  F,  F,    T,    T },
                                    /*4:*/  { F,  F,  F,   null, null},
                                    /*5:*/  { F,  F,  null,null, null}
                                          };
                                          
        
        String algName;
        

        /** 
         *  A new instance is created in case an unacceptable combination of
         *  num wildcards and num rules is found in doUpdate -- in that case
         *  the new instance of the exception is stored in the returned hashtable 
         */
        PoorPerformanceException poorPerfEx = null;
        
        int numRules;
        int avgNumWildcards;
        Updateable rulesInTable;
        public _Table() {}
        /**
         *  The reason for this doUpdate method is this: All the other Action instances
         *  in this class call the doUpdate method on Rules. However, for the TableReadAction 
         *  additional processing must be done by this class, so the doUpdate method is redirected
         *  to this class which, after completing processing, calls the doUpdate method on rules
         */
        public void doUpdate(Hashtable data) {
            Hashtable storedData = (Hashtable)data.get(Action.MAIN_KEY);
            numRules = ((Integer)storedData.get(TableReadAction.NUM_RULES)).intValue();
            avgNumWildcards = ((Integer)storedData.get(TableReadAction.AVG_NUM_WILDCARDS)).intValue();
            rulesInTable = (Updateable)storedData.get(TableReadAction.RULES);
            Boolean consisCheckSwitch = (Boolean)storedData.get(TableReadAction.CONSIS_SWITCH);
            obtainAlgNameAndException();
            Hashtable ht = new Hashtable(11);
            if(algName != null) ht.put(ALGORITHM, algName);
            if(poorPerfEx != null) ht.put(EXCEPTION, poorPerfEx);
            
            if(consisCheckSwitch != null) ht.put(CONSIS_CHECK, 
                                                 consisCheckSwitch);
            
            storedData.put(Action.MAIN_KEY, ht);
            rulesInTable.doUpdate(storedData);
            
        }
        /** 
         *  This method returns true if the table indicates that the number of rules or avg
         *  number of wildcards is too large for tolerable performance. The method is used
         *  by obtainAlgAndConsFlag() and also by Rules to check that a special user request
         *  is safe.
         */  
        public boolean outOfRange() {
            if(numRules < POWERS_OF_2[0]) return false;
            else if(avgNumWildcards > MAX_AVG_NUM_WILDCARDS && 
                    numRules <= POWERS_OF_2[LAST_INDEX]) return false;
            else if(!tableCellNull()) return false;
            else return true;
        }
        private boolean tableCellNull() {
                for(int i = 1; i < POWERS_OF_2.length; ++i) {                
                    if(numRules == POWERS_OF_2[i-1] || //if not, then i must be < POWERS_OF_2.length-1 
                       Util.isBetweenOrEqualOnRight(POWERS_OF_2[i-1], numRules, POWERS_OF_2[i])){
                       
                            //if table entry is null, return true; return false otherwise
                            Boolean tableVal = useManyWildcardAlg[i][avgNumWildcards];
                            return (tableVal == null);
                    }
                }
                return false;
        }
            
        private void obtainAlgNameAndException() {
            //if numrules or num wild cards are such that table has null entry, or values
            //are out of range, set exception and return
            if(outOfRange()) {
                poorPerfEx = new PoorPerformanceException("Unacceptable combination of number of rules ("+
                                                          numRules +
                                                          " and average number of wildcards "+
                                                          avgNumWildcards);
            }
            //if small number of rules, use the many-stars algorithm
            else if(numRules < POWERS_OF_2[0]) {
                algName = Algorithm_manyStars.NAME;
            }
            //if not too many rules and many wildcards, use many-stars algorithm
            else if(avgNumWildcards > MAX_AVG_NUM_WILDCARDS && 
                    numRules < POWERS_OF_2[LAST_INDEX]) {
                        
                algName = Algorithm_manyStars.NAME;
            }
            //if num rules is between 512 and 16384, use table to read off correct algorithm
            else {
                //find appropriate table row
                int rightRow = -1;
                for(int i = 1; i < POWERS_OF_2.length; ++i) {
                    if(numRules == POWERS_OF_2[i-1] || //if not, then i must be < POWERS_OF_2.length-1 
                       Util.isBetweenOrEqualOnRight(POWERS_OF_2[i-1], numRules, POWERS_OF_2[i])){
                        
                        Boolean tableVal = useManyWildcardAlg[i][avgNumWildcards];
                        algName = useManyWildcardAlg[i][avgNumWildcards].booleanValue() ?
                                    Algorithm_manyStars.NAME :
                                    Algorithm_fewStars.NAME;
                        break;
                     }
                }
            }
                    
                
        }
        public int getNumRules() {
            return numRules;
        }
        public void setNumRules(int num) {
            numRules = num;
        }
        public int getAvgNumWildcards() {
            return avgNumWildcards;
        }
        public void setAvgNumWildcards(int num) {
            avgNumWildcards = num;
        }
        public void setRulesInTable(Rules rules) {
            rulesInTable = rules;
        }
        
        
        public Object clone() {
            return new _Table();
        }
    }
            
    /** Unlike most Action instances, this one receives an Updateable instance
     *  that is not the object that ultimately needs to be updated -- instead
     *  the doUpdate call performed here is a request made of _Table to check
     *  against its stored values. The result of this check is then directed
     *  by _Table to the proper destination -- namely, the Rules class.
     */
    public static class TableReadAction extends Action implements Cloneable {
        public static final String NUM_RULES = "num_rules";
        public static final String AVG_NUM_WILDCARDS = "avg_num_wildcards";
        public static final String RULES = "rules";
        public static final String CONSIS_SWITCH  = "consis_switch";
        
        
        public TableReadAction(Updateable ud, Hashtable attributes) {
            super(ud,attributes);
        }
        
        public void setAttributesFromTable(Hashtable attribs) {
            attribute = attribs;
        }
        /** this allows me to keep this class read only, and just use copies as needed */
        protected Object clone() {
            _Table aTable = (_Table)((_Table)ud).clone();
            Hashtable ht = null;
            if(attribute != null) {
                ht = new Hashtable(11);
                Enumeration keys = ((Hashtable)attribute).keys();
                Object key = null;
                while(keys.hasMoreElements()){
                    key = keys.nextElement();
                    ht.put(key, ((Hashtable)attribute).get(key));
                }
            }
            TableReadAction tra = new TableReadAction(aTable,ht);
            return tra;
        }
            
    }
        
        
    public static class AlgAndConsisSetAction extends Action {
        
        public AlgAndConsisSetAction(Updateable ud, Hashtable attributes) {
            super(ud,attributes);
        }
        
        public void setAttributesFromTable(Hashtable attributes) {
            attribute = attributes;
        }
    }        
        
    
    public class _ConcreteInvoker extends Invoker {
        Action manyStarsConOn; 
        Action fewStarsConOn;
        Action fewStarsConOff;
        Action manyStarsConOff;
        TableReadAction readTable;
        
        public _ConcreteInvoker(Updateable updateable) throws PoorPerformanceException {
            super(updateable);
            
        }
        
        public void createActions(Rules rules) {
            
            Hashtable ht = null;
            
            ht = new Hashtable(11);
            ht.put(ALGORITHM, Algorithm_manyStars.NAME);
            ht.put(CONSIS_CHECK, Boolean.TRUE);
            manyStarsConOn = new AlgAndConsisSetAction(rules,ht);
            
            ht = new Hashtable(11);
            ht.put(ALGORITHM, Algorithm_manyStars.NAME);
            ht.put(CONSIS_CHECK,Boolean.FALSE);
            manyStarsConOff = new AlgAndConsisSetAction(rules,ht);
            
            ht = new Hashtable(11);
            ht.put(ALGORITHM, Algorithm_fewStars.NAME);
            ht.put(CONSIS_CHECK,Boolean.TRUE);
            fewStarsConOn = new AlgAndConsisSetAction(rules,ht);
            
            ht = new Hashtable(11);
            ht.put(ALGORITHM, Algorithm_fewStars.NAME);
            ht.put(CONSIS_CHECK,Boolean.FALSE);
            fewStarsConOff = new AlgAndConsisSetAction(rules,ht);
            
            ht = new Hashtable(11);
            //a TableReadAction will only be invoked if consistency checking is switched on
            ht.put(TableReadAction.CONSIS_SWITCH,Boolean.TRUE);
            ht.put(TableReadAction.RULES, rules);
            ht.put(TableReadAction.NUM_RULES, new Integer(rules.getNumRules()));
            ht.put(TableReadAction.AVG_NUM_WILDCARDS, new Integer(rules.getAvgNumStars()));
            readTable = new TableReadAction(new _Table(), ht);
            
        }
            
        
        public void loadRules() throws InconsistentRulesException, 
                                           PoorPerformanceException, 
                                           RuleNotFoundException {
            /** we use internal trickery to ensure direct access to the many-stars algorithm */
            final Algorithm alg = new Algorithm_manyStars();
            final int numConditions = 5;
            final int numRules = 11;
            final int maxNumStars = 3;
            final Boolean consisCheck = Boolean.FALSE;

            alg.init(numConditions,numRules,maxNumStars,consisCheck.booleanValue());
            TableReadAction aTable;
            createActions(outerRules);
            
            
            
            //numRules|consisReq|algReq|consisTrue|algManyWildcards
               //if user does not specify con-check or algorithm, use manyStars and
               //keep con-check turned off
            alg.addRule("*FF**",  manyStarsConOff);
            
               //if user specifies con-check should be switched off, and does not specify
               //algorithm, go with manyStars
            alg.addRule("*TFF*",  manyStarsConOff);
            
               //whether there are many or few rules, if specific algorithm
               //is requested, it must be provided; an exception may be thrown
               //Also, go with no consistency checking if it's not requested
            alg.addRule("*TTTT",  manyStarsConOn);
            alg.addRule("*TTTF",  fewStarsConOn);
            alg.addRule("*TTFT",  manyStarsConOff);
            alg.addRule("*TTFF",  fewStarsConOff);
            alg.addRule("*FT*T",  manyStarsConOff);
            alg.addRule("*FT*F",  fewStarsConOff);
            
               //if no algorithm is requested, and there are few rules use many star algorithm
            alg.addRule("TTFT*",  manyStarsConOn);
                        
               //if consistency-check requested with many rules and no alg specified, consult table
               //we want the TableReadAction to be read only, so we clone it
            aTable =  (TableReadAction)readTable.clone();
            alg.addRule("FTFT*",  aTable);

            //set the algorithm in a new rules instance
            //note that these are the internal rules that determine
            //which scenario of algorithm and consistency checking will be used
            rules = new Rules();
            rules.setNumConditions(numConditions);
            rules.setNumRules(numRules);
            rules.setMaxNumStars(maxNumStars);
            rules.setAlgorithm(alg);
            rules.setConsistencyChecking(consisCheck);
        }            
        
        public void loadConditions() throws IllegalExpressionException, RuleNotFoundException {
            conditions = new Vector(5);
            Boolean consisFlag = outerRules.getConsistencyChecking();
            Algorithm algSelection = outerRules.getAlgorithm();
            //add the num rules < 1000 condition
            Condition numRulesCondition =
                new Condition(outerRules.getNumRules(),
                              Condition.LESS,
                              NUM_RULES_THRESHOLD);
            conditions.addElement(numRulesCondition);
            
            //add the consistencyChecking requested condition
            Condition consisCheckRequestedCondition =
                new Condition(consisFlag != null, true);
            conditions.addElement(consisCheckRequestedCondition);
            
            //add the algorithm requested condition
            Condition algorithmRequestedCondition =
                new Condition(algSelection != null,true);
            conditions.addElement(algorithmRequestedCondition);
            
            //add the consistency check flag value condition
            Condition consistencyFlagSetCondition =
                new Condition((consisFlag != null) && consisFlag.booleanValue(), true);
            conditions.addElement(consistencyFlagSetCondition);
            
            //add the algorithm selection -- test to see if selected algorithm is the
            //many-stars algorithm
            String toBeCompared = (algSelection == null) ? null : algSelection.name();
            Condition algSelectionCondition = 
                new Condition(Algorithm_manyStars.NAME,toBeCompared);
            conditions.addElement(algSelectionCondition);
            
        }
        

    }    
}
