package logic;

import java.util.*;
import java.io.*;

/**
 *  This class is a wrapper for an array of Booleans.
 *  We need a wrapper so that it can be used as a key
 *  in a hashtable. It provides a simple implementation of 
 *  hashCode() for this purpose.
 */
public class BooleanTuple  {
	private Boolean [] tuple;
	
	
	/**
	 *  The constructor accepts a Vector of Booleans. The size of the Vector
	 *  is equal to the number of Conditions-- only MAX_LEVELS_OF_NESTING
	 *  Conditions are allowed (in the original implementation, this maximum
	 *  number is 30). The constuctor converts the Vector to an array 
	 *  of Booleans and caches it.
	 */
	public BooleanTuple(Vector tuple) throws PoorPerformanceException {
		if(tuple != null && tuple.size() > Algorithm_fewStars.MAX_NUMBER_CONDITIONS) {
			throw new PoorPerformanceException("This framework cannot handle more than "+ Algorithm_fewStars.MAX_NUMBER_CONDITIONS+" safely.");
		}
		setTuple(tuple);
	}
	/**
	 *  Two BooleanTuples are equal if they are both nonnull, and their
	 *  underlying instances of Boolean[] have the same length and 
	 *  have entries that have the same booleanValue()'s.
	 */
	public boolean equals(Object ob) {
		if(!(ob instanceof BooleanTuple)) return false;
		BooleanTuple testbt = (BooleanTuple)ob;
		if(ob == null) return false;
		if(tuple == null && testbt.getTuple() == null) return true;
		if(testbt.getTuple() == null) return false;
		if(testbt.size() != size()) return false;
		for(int i = 0; i < size(); ++i) {
			
			Boolean testbtBool = testbt.getTuple()[i];
			Boolean thisBool = tuple[i];
			if(testbtBool.booleanValue() != thisBool.booleanValue()) return false;
		}
		return true;
	}
	public Boolean [] getTuple() {
		return tuple;
	}
	/**
	 *  The algorithm used here is to treat a sequence
	 *  of booleans as a binary sequence, with T=1 and F=0,
	 *  and then treat the binary sequence as an actual
	 *  binary integer. The algorithm simply computes the
	 *  base 10 value of this binary number and returns it.
	 *  An inherent limitation is that the largest allowed int
	 *  is 2^31 -1, so that we can't rely on the algorithm
	 *  for tuples of length > 30.
	 *  Note that this version shows the idea, but is slower
	 *  than necessary since it relies on exponential function.
	 *  The hashCode() function that we use is given below.
	 */
	public int hashCode2() {
		if(tuple == null) return 0;
		int retVal = 0;
		
		for(int i = 0; i < tuple.length; ++i) {
			Boolean b = tuple[i];
			retVal += num(b) * Math.pow(2,i);
		}
		return retVal;
	}
	/**
	 *  The motivation for this method is explained in the
	 *  comments for hashCode2(). This version is faster
	 *  since we avoid using exponentials and make use of the
	 *  convenient representation of powers of 2 in hexadecimal notation
	 */
	public int hashCode() {
		if(tuple == null) return 0;
		int retVal = 0;
		
		for(int i = 0; i < tuple.length; ++i) {
			Boolean b = tuple[i];
			int newVal = num(b);
			if(newVal ==1) {
			  //create hexadecimal representation of the number 2^i 
			  //and add it to accumulator  
			  try {
			    int digit = computeLeadingDigit(i);
			    String zeroes = computeStringOfZeroes(i);
			    String hexVal = "" + digit + zeroes;
			    retVal += Integer.parseInt(hexVal, 16);
			  }
			  catch(NumberFormatException e) {}
			}
		}
		return retVal;
	}
	private String computeStringOfZeroes(int i) {
	    StringBuffer sb = new StringBuffer("");
	    for(int k = 1; k < Algorithm_fewStars.MAX_NUMBER_CONDITIONS; ++k) {
	        if( i >= 4*k) {
	            sb.append("0");
	        }
	        else {
	            return sb.toString();
	        }
	    }
	    return sb.toString();
	}
	private int computeLeadingDigit(int i) {
			    int rem = i % 4;
			    int digit = 0;
			    switch(rem) {
			        case 0:
			            digit = 1;
			            break;
			        case 1:
			            digit = 2;
			            break;
			        case 2:
			            digit = 4;
			            break;
			        case 3:
			            digit = 8;
			            break;
			        default:
			            break;
			    }
			    return digit;
    }	    
	private int num(Boolean b) {
		return b.booleanValue() ? 1 : 0;
	}
    public void setTuple(Boolean [] arr) {
	    tuple = arr;
    }
    public void setTuple(Vector v) {
	    tuple = Util.toBooleanArray(v);
    }
	public int size() {
		return tuple == null ? 0 : tuple.length;
	}
	public String toString() {
	    if(tuple == null) return "null";
	    StringBuffer accum = new StringBuffer();
	    for(int i = 0; i < tuple.length; ++i) {
	        int num = num(tuple[i]);
	        accum.append( ((num == 1) ? "T" : "F") );
	    }
	    return accum.toString();
	}
}