package logic;

import java.util.*;
import java.io.*;

public class BooleanTupleFactory {
    
    /** 
     *  This method generates the array of BooleanTuples that is generated from
     *  the boolean String tf (like "TT**F") -- a String that may contain '*'s.
     *  We first generate all possible boolean strings that don't have stars,
     *  then convert to BooleanTuples. 
     *  <p>
     *  For example, if tf is "TT**F", this method first generates an array
     *  of four non-star boolean Strings as follows:
     *  <pre>
     *     0:  "TTTTF"
     *     1:  "TTTFF"
     *     2:  "TTFTF"
     *     3:  "TTFFF"
     *  </pre>
     *  These four rows are thought of as a 2-d array; the 0th "column" in this
     *  example would be
     *  <pre>
     *     T
     *     T
     *     T
     *     T
     *  </pre>
     *  In order to fill in the '*'s with all possible T's and F's, we use the
     *  truthTable(..) method.
     *  NOTE: Error handling is not being done in case conversion fails for some reason.
     */
    public static void loadTuples(String tf,Algorithm_fewStars alg,Action action) 
                                            throws PoorPerformanceException, InconsistentRulesException {
        Action [] ruleTable = alg.getRuleTable();
        if(tf == null || tf.equals("")) return;
            
        if(tf.length() > Algorithm_fewStars.MAX_NUMBER_CONDITIONS){
            throw new PoorPerformanceException("Can't have more than "+
                   Algorithm_fewStars.MAX_NUMBER_CONDITIONS + " Conditions");    
        }
        int numTuples = countTuples(tf);
        int len = tf.length();
        int numStars = countStars(tf);
        int [] indicesOfStarredCols = {};
        if(numStars > 0) {
            indicesOfStarredCols = computeIndicesOfStarredCols(tf);
        }

        StringBuffer [] tfArr = new StringBuffer[numTuples];
        BooleanTuple [] btArr = new BooleanTuple[numTuples];
        
        for(int col = 0; col < len; ++col) {
            for(int row = 0; row < numTuples; ++row) {
                if(col == 0) tfArr[row] = new StringBuffer("");             
                char c = tf.charAt(col);
                
                tfArr[row].append(obtainBoolChar(col, row, c, indicesOfStarredCols));
            }
        }
        BooleanTuple booltup = null;
        for(int row = 0; row < numTuples; ++row) {
            booltup = getTuple(tfArr[row].toString());
            int hash = booltup.hashCode();
            if(alg.getConsistencyChecking()) {
                checkConsistency(action, hash, booltup, ruleTable);
            }
            ruleTable[hash] = action;
        }
        alg.setRuleTable(ruleTable);
    }
    
    
    //------------ PRIVATE METHODS --------------//
    
    /**
     *  If character c is just 'T' or 'F', this method just returns c. If c is a star, however,
     *  then this method converts the column index of this star to the index of the
     *  array indicesOfStarredCols that contains it, and then calls truthTable(..) to 
     *  determine whether, for the  given column and row, the value should be 'T' or 'F'
     *  NOTE: No error handling is done in case incorrect data is being passed around.
     */
    private static char obtainBoolChar(int col, int row, char c, int[] indicesOfStarredCols) {
        switch(c) {
            case 'T':
            case 't':
                return 'T';

            case 'F':
            case 'f':
                return 'F';

            case '*':
                //find index of starred col index in the indicesOfStarredCols array
                int adjustedColNum = Util.findIndex(col, indicesOfStarredCols);
                
                //call truthTable to determine appropriate truth value for this row and column
                int booleanAsNumeric = 
                    //truthTable(adjustedColNum, row, indicesOfStarredCols.length);
                    Util.TRUTH_TABLES[indicesOfStarredCols.length][adjustedColNum][row];
                    
                //convert to 'T' or 'F'
                return (booleanAsNumeric == 0) ? 'T' : 'F';

            default:
                break;
        }
        return 'F';
    }
    private static void checkConsistency(Action action, int hash, BooleanTuple booltup, Action[] ruleTable) throws InconsistentRulesException {
       Action currentAction = ruleTable[hash];
       if(currentAction != null && !currentAction.equals(action)) {
           throw new InconsistentRulesException("The BooleanTuple "+booltup+" is being associated "+
                                                "with two different actions.");
       } 
    }
    
    
    
    /**
     *  Counts the number of '*'s that occur in s
     */
    private static int countStars(String s) {
        int count = 0;
        if(s == null) return count;
        for(int i = 0; i < s.length(); ++i) {
            if(s.charAt(i) == '*') ++count;
        }
        return count;
    }
    
    /**
     *  Places the positions of '*' in s into an array. Returns an empty (non-null)
     *  array if there are no '*'s.
     */
    private static int [] computeIndicesOfStarredCols(String s) {
        int arraySize = countStars(s);
        int [] arrayOfIndices = new int[arraySize];
        int currIndex = 0;
        for(int i = 0; i < s.length(); ++i) {
            if(s.charAt(i) == '*') {
                arrayOfIndices[currIndex] = i;
                ++currIndex;
            }
        }
        return arrayOfIndices;
    }
    
    /**
     *  Counts the number of boolean strings (like "TFFTF") that will be derived
     *  from s. If s has no stars, the return value is 1. If 1 star, the value is
     *  2. In general, the return value is 2^num_stars
     */
    private static int countTuples(String s) {
        return (int)Math.pow(2,countStars(s));
    }
    
    /**
     *  Converts 'T' and 'F' to the corresponding Booleans
     */
    private static Boolean getBoolean(char c) {
        if(c == 't' || c == 'T') return Boolean.TRUE;
        return Boolean.FALSE;
    }
    /**
     *  This method returns a vector of Booleans, wrapped in a BooleanTuple, that
     *  corresponds to a boolean String like "TTFFF". Used by getTuples(..).
     *  This method does not handle Strings that contain stars, and assumes
     *  that all characters in tf are either 'T' or 'F'
     */
    private static BooleanTuple getTuple(String tf) throws PoorPerformanceException {
        
        if(tf == null || tf.equals("")) return null;
        if(tf.length() > Algorithm_fewStars.MAX_NUMBER_CONDITIONS){
            throw new PoorPerformanceException("Can't have more than "+
                   Algorithm_fewStars.MAX_NUMBER_CONDITIONS + " Conditions");    
        }
        int len = tf.length();
        Vector v = new Vector(len);
        
        for(int i = 0; i < len; ++i) {
            char c = tf.charAt(i);
            Boolean b = getBoolean(c);
            v.addElement(b);
        }
        return new BooleanTuple(v);
    }
    
    
}
