package logic;

import java.util.*;
/**
 *  This class abstracts the notion of a "condition" in branching logic.
 *  It provides constructors for a variety of simple kinds of conditions
 *  such as comparing primitives with =, <, or <=. It also provides
 *  a constructor for testing whether objects are equal, and a more
 *  general constructor, which accepts a Hashtable, for more complicated conditions.
 */
public class Condition {
		final int BOOLEAN = 0;
		final int INTEGER = 1;
		final int FLOAT = 2;
		final int DOUBLE = 3;
		final int OBJECT = 4;
		int status=-1;
		Hashtable data;
		Object ob1, ob2;
		boolean b1, b2;
		int i1, i2;
		float f1, f2;
		double d1,d2;
		int op=-1;
		public static final int LESS = 0;
		public static final int LESS_EQUAL = 1;
		public static final int EQUAL = 2;
		    
		
	    /**
	     *  This constructor compares two int's using the comparison named by oper.
	     *  The default evaluate() method should be used.
	     */
		public Condition(int i1, int oper, int i2) throws IllegalExpressionException {
												
				if(badOperator(oper)){
					throw new IllegalExpressionException("Illegal operator: oper ="+oper);
				}
				this.i1 = i1;
				this.op = oper;
				this.i2 = i2;
				status = INTEGER;	 
		}
		
	    /**
	     *  This constructor compares two float's using the comparison named by oper.
	     *  The default evaluate() method should be used.
	     */
		public Condition(float f1, int oper, float f2)	throws IllegalExpressionException {
												
				if(badOperator(oper)){
					throw new IllegalExpressionException("Illegal operator: oper ="+oper);
				}
				this.f1 = f1;
				this.op = oper;
				this.f2 = f2;
				status = FLOAT;	   
		}

	    /**
	     *  This constructor compares two doubles using the comparison named by oper.
	     *  The default evaluate() method should be used.
	     */
		public Condition(double d1, int oper, double d2) throws IllegalExpressionException {
												
				if(badOperator(oper)){
					throw new IllegalExpressionException("Illegal operator: oper ="+oper);
				}
				this.d1 = d1;
				this.op = oper;
				this.d2 = d2;
				status = DOUBLE;	
		}
		
	    /**
	     *  This constructor compares two booleans to see if they are the same.
	     *  The default evaluate() method should be used.
	     */
		public Condition(boolean b1, boolean b2) {
				this.b1 = b1;
				this.b2 = b2;
				status = BOOLEAN;	 
		}
		
		/**
		 *  This version of Condition accepts a Hashtable that must contain
		 *  all the data that is necessary to implement the evaluate() method --
		 *  the evaluate() method must be overridden if this constructor is used.
		 */
		public Condition(Hashtable data) {
				this.data = data;
		}
		
		/**
		 *  This constructor is used when two objects need to be compared.
		 *  Usually, the default equals(Object) method provided by Object
		 *  should be overridden in the class that is the template for
		 *  the arguments ob1 and ob2.
		 */
		public Condition(Object ob1, Object ob2) {
				  
				this.ob1 = ob1;
				this.ob2 = ob2;
				status = OBJECT;
		 }
		/** If the Hashtable version of the constructor is used, this
		 *	method must be overridden; otherwise the present version
		 *	of evaluate should work without modification
		 */
		public Boolean evaluate() throws DataNotFoundException {

			switch(status) {
				case BOOLEAN:
					return evaluateBoolean();
				case INTEGER:
					return evaluateInteger();
				case FLOAT:
					return evaluateFloat();
				case DOUBLE:
					return evaluateDouble();
				case OBJECT:
					return evaluateObject();
				default:
								
			}
			return Boolean.FALSE;
		}
		private Boolean evaluateBoolean() {
			return new Boolean( b1 == b2 );
		}
		/** Note that the equals() method of Object should have been overridden
		 *   in order for this method to work as desired.
		 */
		private Boolean evaluateObject() { 
			return new Boolean(ob1.equals(ob2));
		}
		private Boolean evaluateDouble() {
			switch(op) {
				case LESS:
					return new Boolean(d1 < d2);
				case LESS_EQUAL:
					return new Boolean(d1 <= d2);
				case EQUAL:
					return new Boolean(d1 == d2);
				default:
					return null;
			 }
		}
		private Boolean evaluateFloat() {
			switch(op) {
				case LESS:
					return new Boolean(f1 < f2);
				case LESS_EQUAL:
					return new Boolean(f1 <= f2);
				case EQUAL:
					return new Boolean(f1 == f2);
				default:
					return null;
			 }
			 
				
		}

		private Boolean evaluateInteger() {
			switch(op) {
				case LESS:
					return new Boolean(i1 < i2);
				case LESS_EQUAL:
					return new Boolean(i1 <= i2);
				case EQUAL:
					return new Boolean(i1 == i2);
				default:
					return null;
			 }
			 
				
		}

		public Hashtable getData() {
			return data;
		}
		
		/** returns true if the argument op is not one of the legal operators */
		private boolean badOperator(int oper) {
			return (oper != LESS && 
					oper != LESS_EQUAL && 
					oper != EQUAL);
		}
		
		public String toString() {
		    StringBuffer sb = new StringBuffer("[Condition: ");
		    sb.append("status=");
		    switch(status) {
		        case(BOOLEAN) :
		            sb.append("boolean,");
		            break;
		        case(INTEGER) :
		            sb.append("integer,");
		            break;
		        case(FLOAT) :
		            sb.append("float,");
		            break;
		        case(DOUBLE) :
		            sb.append("double,");
		            break;
		        case(OBJECT) :
		            sb.append("object,");
		            break;
		        default :
		            sb.append("hashtable,");
		            break;
		    }
		    sb.append("op=");
		    if(1 <= status && status <= 3) {
		        if(op==LESS) sb.append("less,");
		        if(op==EQUAL) sb.append("equal,");
		        if(op==LESS_EQUAL) sb.append("less_equal,");
		    }
		    else if(status==OBJECT || status == BOOLEAN) {
		        sb.append("equal,");
		    }
		    else {
		        sb.append("none,");
		    }
		    if(0 <= status && status <= 4) {
		        switch(status) {
    		        case(BOOLEAN) :
	    	            sb.append("1st=");
	    	            sb.append(b1);
	    	            sb.append(",");
	    	            sb.append("2nd=");
	    	            sb.append(b2);
	    	            sb.append("]");
		                break;
		            case(INTEGER) :
	    	            sb.append("1st=");
	    	            sb.append(i1);
	    	            sb.append(",");
	    	            sb.append("2nd=");
	    	            sb.append(i2);
	    	            sb.append("]");
		                break;
		            case(FLOAT) :
	    	            sb.append("1st=");
	    	            sb.append(f1);
	    	            sb.append(",");
	    	            sb.append("2nd=");
	    	            sb.append(f2);
	    	            sb.append("]");
		                break;
		            case(DOUBLE) :
	    	            sb.append("1st=");
	    	            sb.append(d1);
	    	            sb.append(",");
	    	            sb.append("2nd=");
	    	            sb.append(d2);
	    	            sb.append("]");
		                break;
		            case(OBJECT) :
	    	            sb.append("1st=");
	    	            sb.append(ob1);
	    	            sb.append(",");
	    	            sb.append("2nd=");
	    	            sb.append(ob2);
	    	            sb.append("]");
		                break;
		            default :
		                break;
		        }
		    }
		    //hashtable is being used
		    else {
		        sb.append("objects=(hashtable displayed below]");
		        sb.append(System.getProperty("line.separator"));
		        sb.append("   ");
		        sb.append(data);
		    }
		    
		    return sb.toString();
		}
		
}
