package logic;

import java.util.*;
import logic.*;

/**
 *  This class is the interface to the whole If-Then-Else framework.
 *  It is responsible for retrieving the Conditions and Rules 
 *  created by user classes and delivering them to the rules
 *  engine (the IfThenElse class) for processing. The result
 *  of this processing shows up in the Updateable instance that
 *  is passed in to the main constructor.
 */
abstract public class Invoker  {
    protected Rules rules;
    protected Vector conditions;
    protected Vector actions;
    protected Updateable ud;

public Invoker() {
}

public Invoker(Updateable u) throws PoorPerformanceException {
    this.ud = u;
    //init();

}
/**
 * This method fires off the if-then-else logic.
 * It requires that the init() method has already been invoked.
 * It should not have to be overridden.
 * @author Paul Corazza
 */
public void execute() throws DataNotFoundException, 
                             PoorPerformanceException, 
                             IllegalExpressionException, 
                             RuleNotFoundException,
                             InconsistentRulesException {
    
        loadConditions();
        new IfThenElse(conditions, rules);
}
public Vector getConditions() {
    return conditions;
}
public Rules getRules() {
    return rules;
}
/**
 * This method should be called at "startup" -- some time
 * before if-then-else logic needs to be performed.
 * The loadRules() method should always be called here;
 * if more initialization is required, init() can be overridden.
 * @author Paul Corazza
 */
public void init() throws PoorPerformanceException, 
                          RuleNotFoundException,
                          InconsistentRulesException {
    loadRules();
}
abstract protected void loadConditions() throws IllegalExpressionException,
                                                RuleNotFoundException;
abstract protected void loadRules() throws InconsistentRulesException, 
                                           PoorPerformanceException, 
                                           RuleNotFoundException;
}
