package logic;

import java.util.*;

/**
 *  This class encapsulates the notion of a set of rules that
 *  match a sequence of booleans with an action.
 *  The underlying container of these match-ups is a Hashtable.
 *  The Rules class provides an interface to this underlying
 *  data structure with addRule(..) and lookUp(..)
 */
public class Rules implements Updateable {
    
	
	/** this is the actual number of calls to addRule--not the
	 *  number of rules obtained by resolving wildcards
	 */
	private int numRules;
	private int numConditions;
	private int maxNumStars;
	private int avgNumStars;
	
	
	/** no public setter allowed */
	private Algorithm algorithm;
	
	/** no public setter allowed */
	private Boolean consistencyChecking;
	
	/** When set to true, this flag ensures performance checking (particularly
	 *  when consistencyChecking is switched on)
	 *  This flag can be switched off (not recommended) by using the most
	 *  complicated of the Rules constructors
	 */
	private boolean safetyOn = true;
	
	/** this may be set by doUpdate in case algorithm selection runs into
	 *  disallowed combinations
	 */
	private PoorPerformanceException exception;
	
	/** Default constructor for internal use only */
	Rules() {}
	    
		
	/**
	 *  Precondition: The user has set all requested options:
	 *  consistency checking, choice of algorithm, or neither
	 *  Typically, Rules should be instantiated in the loadRules
	 *  method of a concrete subclass of Invoker. 
	 */
	public Rules(int numConditions, int numRules, int maxNumStars)
	             throws PoorPerformanceException {
	                
	    this(numConditions,numRules, maxNumStars, null, null);
	        
	}
	/** 
	 *  This version of the constructor is used if either
	 *  consistency checking or a specific algorithm are to
	 *  be selected. Note that there is no way to set these
	 *  selections after the constructor has been invoked --
	 *  even if it could be done, it would be too late because
	 *  the values for consistency checking and algorithm will
	 *  have already been made.
	 */
	public Rules(int numConditions, int numRules, int maxNumStars, Algorithm alg, Boolean consisCheck) 
	             throws PoorPerformanceException {
	    this.numConditions = numConditions;
	    this.numRules = numRules;
	    this.maxNumStars = maxNumStars;
	    this.algorithm = alg;
	    this.consistencyChecking = consisCheck;
	    computeAvgNumStars();
	    setAlgAndConsisCheckFlag();
	    
	}
	/**
	 *  This version of the constructor is like the last one, but allows you to switch
	 *  off safetyOn, which is used in the abortIfDangerous() safety check
	 */
	public Rules(int numConditions, int numRules, 
	             int maxNumStars, Algorithm alg, 
	             Boolean consisCheck, boolean safetyOn) 
	                throws PoorPerformanceException {
	                    
	    this(numConditions,numRules,maxNumStars,alg,consisCheck);
	    this.safetyOn = safetyOn;
	}
	 
	/**
	 *  There are two reasons why an exception could be thrown here. 
	 *  First, work in AlgorithmAndConsisCheckSelector may have 
	 *  resulted in a PoorPerformanceException, which was set by doUpdate().
	 *  Second, table in AlgorithmAndConsisCheckSelector may not have  
	 *  been consulted because the user may have requested specific options
	 *  that were honored, and that, according to the table, 
	 *  form an unworkable combination.
	 */
	private void abortIfDangerous() throws PoorPerformanceException {
	    //there's no way to avert this check -- it means that an appropriate
	    //algorithm could not be found in the AlgorithmAndConsisCheckSelector
	    if(exception != null) throw exception;
	    
	    //this block of safety checking can be omitted by switching safetyOn off
	    if(safetyOn && consistencyChecking != null && consistencyChecking.booleanValue()) {
	        AlgorithmAndConsisCheckSelector._Table table = 
	            new AlgorithmAndConsisCheckSelector._Table();
	        table.setNumRules(numRules);
	        table.setAvgNumWildcards(avgNumStars);
	        table.setRulesInTable(this);
	        if(table.outOfRange()) {
	            String msg = "PoorPerformanceException: "+
	                         "This combination of "+
	                         numRules +
	                         " rules and "+
	                         avgNumStars+
	                         " stars on average"+
	                         " would result in poor performance when consistency-checking is turned on.";
	            throw new PoorPerformanceException(msg);
	        }
	    }
	}
	
	private void computeAvgNumStars() {
	    double maxRulesOverNumRules = Math.pow(2,numConditions)/numRules;
	    avgNumStars = Util.logBase2AsInt(maxRulesOverNumRules);
	}
	
	public int getAvgNumStars() {
	    return avgNumStars;
	}
	public int getNumRules() {
	    return numRules;
	}
	
	public Boolean getConsistencyChecking() {
	    return consistencyChecking;
	}
	
	private void setAlgAndConsisCheckFlag() throws PoorPerformanceException {
	
	    try {
	        AlgorithmAndConsisCheckSelector selector = 
	            new AlgorithmAndConsisCheckSelector(this);
	        Invoker selectionInvoker = selector.new _ConcreteInvoker((Updateable)this);
	        selectionInvoker.init();
	        
	        // this selects and sets the algorithm and consistency checking flag
	        selectionInvoker.execute();
	        abortIfDangerous();
	    }
	    
	    catch(PoorPerformanceException e) {
	        throw e;
	    }
	    catch(Exception e) {
	        throw new PoorPerformanceException("PoorPerformanceException: Unable to obtain appropriate algorithm for this combination of rules and wildcards.");    
	    }
	}
	    
	        
	public void addRule(String boolStr, Action action) throws InconsistentRulesException, 
	                                                          PoorPerformanceException,
	                                                          RuleNotFoundException {
	    algorithm.addRule(boolStr,action);
	}
	public Action lookUp(Vector results) throws PoorPerformanceException, 
	                                            InconsistentRulesException,
	                                            RuleNotFoundException {
	    return algorithm.lookUp(results);
	}
	
	/** This is the implementation of Updateable, used to select algorithm,
	 *  set consistency checking flag, and set exception
	 */
	public void doUpdate(Hashtable ht) {
	    Hashtable data = (Hashtable)ht.get(Action.MAIN_KEY);
	    String algKey = AlgorithmAndConsisCheckSelector.ALGORITHM;
	    String consisKey = AlgorithmAndConsisCheckSelector.CONSIS_CHECK;
	    String exceptionKey = AlgorithmAndConsisCheckSelector.EXCEPTION;
	    
	    if(data.containsKey(consisKey)) {
	        consistencyChecking = (Boolean)data.get(consisKey);
	    }
	    if(data.containsKey(algKey)) {
	        String algName = (String)data.get(algKey);
	        algorithm = (Algorithm)Util.getInstance(algName);
	        
	    }
	    if(data.containsKey(exceptionKey)) {
	        exception = (PoorPerformanceException)data.get(exceptionKey);
	    }
	    
	    //either algorithm or exception must be non-null
	    if(algorithm == null && exception == null) {
	        exception = new PoorPerformanceException("Unable to select appropriate algorithm");
	    }
	    if(algorithm != null) {
	        algorithm.init(numConditions, 
	                       numRules, 
	                       maxNumStars,
	                       consistencyChecking.booleanValue());
	    }
	    
	}
	public Algorithm getAlgorithm() {
	    return algorithm;
	}
	
	// these setters are for internal use only
	void setNumRules(int val) {
	    numRules = val;
	}
	void setNumConditions(int val) {
	    numConditions = val;
	}
	void setMaxNumStars(int val) {
	    maxNumStars = val;
	}
	void setAlgorithm(Algorithm alg) {
	    algorithm = alg;
	}
	void setConsistencyChecking(Boolean b) {
	    consistencyChecking = b;
	}
	

}